% Copyright (C) 2012  Arno Onken
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% -------------------------------------------------------------------------
% Discrete bivariate maximum entropy distribution where the marginal
% distributions and the correlation coefficient are constrained (see Pasha
% and Mansoury, Applied Mathematical Sciences, 2008).
%
% Arguments:
%  margin1     - Probability mass function vector of margin 1
%  margin2     - Probability mass function vector of margin 2
%  correlation - Correlation coefficient
%
% Returns:
%  p_me   - Probability mass function of joint maximum entropy distribution
%  f1     - Values f1 of exponential representation
%  f1     - Values f2 of exponential representation
%  lambda - lambda value of exponential representation
% -------------------------------------------------------------------------
function [p_me, f1, f2, lambda] = discmaxentestimate (margin1, margin2, correlation)

    % Check arguments
    if (nargin ~= 3)
        error ('usage: [p_me, f1, f2, lambda] = discmaxentestimate (margin1, margin2, correlation)');
    end

    if (~isvector (margin1))
        error ('discmaxentestimate: margin1 must be a pmf vector of margin 1');    
    end
    margin1 = margin1(:)' ./ sum (margin1);

    if (~isvector (margin2))
        error ('discmaxentestimate: margin2 must be a pmf vector of margin 2');    
    end
    margin2 = margin2(:)' ./ sum (margin2);

    if (~isscalar (correlation) || correlation < -1 || correlation > 1)
        error ('discmaxentestimate: correlation must be a scalar in [-1, 1]');
    end
    
    % Transform correlation to expectation of the variable product
    n1 = length (margin1);
    n2 = length (margin2);
    % Expectations
    mu1 = dot (margin1, 0:n1-1);
    mu2 = dot (margin2, 0:n2-1);
    % Standard deviations
    sigma1 = sqrt (dot (margin1, ((0:n1-1) - mu1).^2));
    sigma2 = sqrt (dot (margin2, ((0:n2-1) - mu2).^2));

    % Function to solve
    valfun = @(x)maxent_val (x, margin1, margin2, correlation*(sigma1*sigma2) + mu1*mu2);
    % Start with independent distribution
    x = fsolve (valfun, [margin1 margin2 0], optimset ('Display', 'off'));
    
    f1 = x(1:n1);
    f2 = x((n1+1):(n1+n2));
    lambda = x(end);
    % Compute maximum entropy distribution
    p_me = (f1' * f2) .* exp (lambda .*  (0:n1-1)' * (0:n2-1));
    % Make sure that all values are non-negative
    p_me(p_me < 0) = 0;
    % Renormalize distribution
    p_me = p_me ./ sum (p_me(:));
end

% Error function for finding the discrete bivariate maximum entropy
% distribution with marginals and correlation coefficient constraints
function y = maxent_val (x, margin1, margin2, e_12)
    n1 = length (margin1);
    n2 = length (margin2);

    % Separate input
    f1 = x(1:n1);
    f2 = x((n1+1):(n1+n2));
    lambda = x(end);

    % Output vector
    y = zeros (size (x));
    y(1:n1) = f1 .* (f2 * exp (lambda .* (0:(n2-1))' * (0:(n1-1)))) - margin1;
    y((n1+1):(n1+n2)) = f2 .* (f1 * exp (lambda .* (0:(n1-1))' * (0:(n2-1)))) - margin2;
    
    % Calculate product expectation
    s = sum (sum (((0:(n1-1))' * (0:(n2-1))) .* (f1' * f2) .* exp (lambda .* (0:(n1-1))' * (0:(n2-1)))));
    y(end) = s - e_12;
end
