% Copyright (C) 2012  Arno Onken
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% -------------------------------------------------------------------------
% Test for linear correlation maximum entropy between x_cond and y_cond,
% assuming that x_cond and y_cond are cell arrays. with conditional Poisson
% distributed integer vectors. The mutual information is used as a
% divergence measure.
%
% Arguments:
%  x_cond - Cell array of conditional integer random values
%  y_cond - Cell array of conditional integer random values of the same
%           size as x_cond; x_cond{s} and y_cond{s} are sampled with some
%           probability p(x_cond{s},y_cond{s}|s) for a given condition s
%  p_cond - Vector of probabilities p(s) or counts of all conditions
%  alpha  - Significance level (default alpha = 0.05)
%  n_mc   - Number of Monte Carlo samples (default n_mc = 1000)
%
% Returns:
%  h      - 1 indicates rejection of the linear correlation maximum entropy
%           hypothesis at the specified significance level; 0 otherwise
%  p      - p-value at the specified significance level
% -------------------------------------------------------------------------
function [h, p] = mcmaxentpoissmitest (x_cond, y_cond, p_cond, alpha, n_mc)
    % Check arguments
    if (nargin < 3)
        error ('usage: [h, p] = mcmaxentpoissmitest (x_cond, y_cond, p_cond [, alpha, n_mc])');
    end

    if (nargin < 4 || isempty (alpha))
        alpha = 0.05;
    end

    if (nargin < 5 || isempty (n_mc))
        n_mc = 1000;
    end

    if (length (x_cond) ~= length (y_cond) || ~iscell (x_cond) || ~iscell (y_cond))
        error ('mcmaxentpoissmitest: x_cond and y_cond must be cell arrays of the same length');
    end
    
    if (~isvector (p_cond) || length (p_cond) ~= length (x_cond))
        error ('mcmaxentpoissmitest: p_s must be a vector of the same length as x_cond');
    end
    p_cond = p_cond(:) ./ sum (p_cond);

    if (~isscalar (alpha) || alpha < 0 || alpha > 1)
        error ('mcmaxentpoissmitest: alpha must be a scalar in [0, 1]');
    end

    if (~isscalar (n_mc) || round (n_mc) ~= n_mc || n_mc < 1)
        error ('mcmaxentpoissmitest: n_mc must be a positive integer');
    end

    % Number of conditions
    n_c = length (p_cond);
    % Number of samples
    n = zeros (n_c, 1);
    % Empirical distributions
    pmf_emp = cell (n_c, 1);
    % Determine maximum spike counts
    max1 = zeros (1, n_c);
    max2 = zeros (1, n_c);
    for c = 1:n_c
        max1(c) = max (x_cond{c}) + 1;
        max2(c) = max (y_cond{c}) + 1;
    end
    % Generate empirical distributions and construct initial theta
    theta_init = zeros (3 * n_c, 1);
    lambda1_emp = zeros (n_c, 1);
    lambda2_emp = zeros (n_c, 1);
    % Minimum sizes for reference distributions
    sz = zeros (n_c, 2);
    for c = 1:n_c
        n(c) = length (x_cond{c});
        % Generate contingency table
        cont = zeros (max (max1), max (max2));
        for l = 1:n(c)
            cont(x_cond{c}(l)+1, y_cond{c}(l)+1) = cont(x_cond{c}(l)+1, y_cond{c}(l)+1) + 1;
        end
        pmf_emp{c} = cont ./ n(c);

        % Set initial theta
        [lambda1_emp(c) lambda2_emp(c) correlation_emp] = constraints (pmf_emp{c});
        if (lambda1_emp(c) <= 0 || lambda2_emp(c) <= 0)
            error ('mcmaxentpoissmitest: empirical rates must be positive');
        end
        theta_init((c-1)*3 + 1) = lambda1_emp(c);
        theta_init((c-1)*3 + 2) = lambda2_emp(c);
        theta_init(c*3) = correlation_emp;
        % Minimum size vector for reference distributions
        sz(c, :) = max (size (cont), poissinv (1-1e-4, [lambda1_emp(c) lambda2_emp(c)]) + 1);
    end
    % Mutual information of empirical distribution
    mi_emp = mi (pmf_emp, p_cond);

    %% Search supremum of survival function
    % Survival function
    s = @(theta) survival (theta, p_cond, mi_emp, max (sz), n, n_mc);
    % Bounds of parameters
    lb = eps * ones (size (theta_init));
    lb((1:n_c) .* 3) = -1;
    ub = [inf(1, n_c); inf(1, n_c); ones(1, n_c)];
    ub = ub(:);
    % Set initial temperature to speed up optimization
    t_init = sqrt (max ([lambda1_emp; lambda2_emp])) / sqrt (min (n));
    % Apply simulated annealing
    g = sup (s, theta_init, lb, ub, t_init);

    % Monte Carlo finite sample correction
    p = (n_mc * g + 1) / (n_mc + 1);
    h = p < alpha;
end

% Survival function of the test statistic
function g = survival (theta, p_cond, mi_emp, sz, n, n_mc)
    n_c = length (n);
    % Extract arguments
    lambda1 = theta(((1:n_c)-1)*3 + 1);
    lambda2 = theta(((1:n_c)-1)*3 + 2);
    correlation = theta((1:n_c)*3);
    correlation(correlation < -1) = -1;
    correlation(correlation > 1) = 1;
    max1 = max ([(sz(1) - 1); poissinv(1-1e-4, lambda1)]);
    max2 = max ([(sz(2) - 1); poissinv(1-1e-4, lambda2)]);
    pmf_me = cell (n_c, 1);
    for c = 1:n_c
        margin1 = poisspdf (0:max1, lambda1(c));
        margin2 = poisspdf (0:max2, lambda2(c));
        % Maximum entropy model for fixed marginals and correlation
        pmf_me{c} = discmaxentestimate (margin1, margin2, correlation(c));
    end

    % Divergence measure: Mutual information difference
    test_stat = abs (mi_emp - mi (pmf_me, p_cond));

    %% Find critical region via Monte Carlo sampling
    sample_stat = zeros (n_mc, 1);
    sample_pmf_emp = cell (n_c, 1);
    for i = 1:n_mc
        for c = 1:n_c
            % Draw a sample from the multinomial distribution
            sample_pmf_emp{c} = reshape (mnrnd (n(c), pmf_me{c}(:)), [max1 max2]+1) ./ n(c);
        end
        sample_stat(i) = abs (mi (sample_pmf_emp, p_cond) - mi (pmf_me, p_cond));
    end
    % Correct for ties
    n_ties = sum (test_stat == sample_stat);
    u = rand (n_ties + 1, 1);
    % Estimate survival function
    g = (sum (test_stat < sample_stat) + sum (u(end) >= u(1:n_ties))) / n_mc;
end

% Constraints for the maximum entropy distribution
function [lambda1 lambda2 correlation] = constraints (p)
    [n1, n2] = size (p);
    % Marginal distributions
    margin1 = sum (p, 2)';
    margin2 = sum (p, 1);
    % Expectations
    lambda1 = dot (margin1, 0:n1-1);
    lambda2 = dot (margin2, 0:n2-1);
    % Variances
    sigma1 = dot (margin1, ((0:n1-1)-lambda1).^2);
    sigma2 = dot (margin2, ((0:n2-1)-lambda2).^2);
    % Correlation coefficient
    if (sigma1 == 0 || sigma2 == 0)
        correlation = 0;
    else
        correlation = (sum (sum (p .* ((0:n1-1)' * (0:n2-1)))) - lambda1*lambda2) ./ sqrt (sigma1*sigma2);
    end
end

% Mutual information
function i = mi (pmf_cond, p_cond)
    n_c = length (p_cond);
    % Conditional entropy
    h_cond = 0;
    % Unconditional probability mass function
    pmf = zeros (size (pmf_cond{1}));
    for c = 1:n_c
        pmf_cond_c = pmf_cond{c}(pmf_cond{c} > 0);
        h_cond = h_cond - p_cond(c) .* sum (log2 (pmf_cond_c) .* pmf_cond_c);
        pmf = pmf + (p_cond(c) .* pmf_cond{c});
    end
    pmf = pmf(pmf > 0);
    i = -sum (log2 (pmf) .* pmf) - h_cond;
end
