% Copyright (C) 2012  Arno Onken
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% -------------------------------------------------------------------------
% Test for linear correlation maximum entropy between x and y, assuming
% that the elements of x and y are Poisson distributed. The entropy
% difference is used as a divergence measure.
%
% Arguments:
%  x      - Vector of integer random values
%  y      - Vector of integer random values of the same size as x
%  alpha  - Significance level (default alpha = 0.05)
%  n_mc   - Number of Monte Carlo samples (default n_mc = 1000)
%
% Returns:
%  h      - 1 indicates rejection of the linear correlation maximum entropy
%           hypothesis at the specified significance level; 0 otherwise
%  p      - p-value at the specified significance level
% -------------------------------------------------------------------------
function [h, p] = mcmaxentpoisstest (x, y, alpha, n_mc)
    % Check arguments
    if (nargin < 2)
        error ('usage: [h, p] = mcmaxentpoisstest (x, y [, alpha, n_mc])');
    end

    if (nargin < 3 || isempty (alpha))
        alpha = 0.05;
    end

    if (nargin < 4 || isempty (n_mc))
        n_mc = 1000;
    end

    if (length (x) ~= length (y))
        error ('mcmaxentpoisstest: x must have the same length as y');    
    end

    if (~isscalar (alpha) || alpha < 0 || alpha > 1)
        error ('mcmaxentpoisstest: alpha must be a scalar in [0, 1]');
    end

    if (~isscalar (n_mc) || round (n_mc) ~= n_mc || n_mc < 1)
        error ('mcmaxentpoisstest: n_mc must be a positive integer');
    end

    n = length (x);
    % Generate contingency table
    cont = zeros (max (x) + 1, max (y) + 1);
    for i = 1:n
        cont(x(i)+1, y(i)+1) = cont(x(i)+1, y(i)+1) + 1;
    end

    %% Maximum entropy model for fixed marginals and correlation
    p_emp = cont / n;
    [lambda1_emp lambda2_emp correlation_emp] = constraints (p_emp);
    if (lambda1_emp <= 0 || lambda2_emp <= 0)
        error ('mcmaxentpoisstest: empirical rates must be positive');
    end
    % Entropy of empirical distribution
    h_emp = entropy (cont(:) / n);

    %% Search supremum of survival function
    % Minimum size vector for reference distributions
    sz = max (size (cont), poissinv (1-1e-4, [lambda1_emp lambda2_emp]) + 1);
    % Survival function
    s = @(theta) survival (theta, h_emp, sz, n, n_mc);
    % Initial theta
    theta_init = [lambda1_emp lambda2_emp correlation_emp];
    % Bounds of parameters
    lb = [0 0 -1];
    ub = [inf inf 1];
    % Set initial temperature to speed up optimization
    t_init = max (std (x), std (y)) / sqrt (n);
    % Apply simulated annealing
    g = sup (s, theta_init, lb, ub, t_init);

    % Monte Carlo finite sample correction
    p = (n_mc * g + 1) / (n_mc + 1);
    h = p < alpha;
end

% Survival function of the test statistic
function g = survival (theta, h_emp, sz, n, n_mc)
    % Extract arguments
    lambda1 = theta(1);
    lambda2 = theta(2);
    correlation = theta(3);
    correlation(correlation < -1) = -1;
    correlation(correlation > 1) = 1;
    margin1 = poisspdf (0:max ([(sz(1) - 1) poissinv(1-1e-4, lambda1)]), lambda1);
    margin2 = poisspdf (0:max ([(sz(2) - 1) poissinv(1-1e-4, lambda2)]), lambda2);

    % Find corresponding maximum entropy distribution
    pmf_me = discmaxentestimate (margin1, margin2, correlation);
    pmf_me = pmf_me(:);
    h_me = entropy (pmf_me);

    % Calculate reference statistic
    test_stat = abs (h_emp - h_me);

    %% Find critical region via Monte Carlo sampling
    % Draw samples from the multinomial distribution
    sample = mnrnd (n, pmf_me, n_mc)' ./ n;
    % Calculate test statistics
    sample_stat = abs (entropy (sample) - h_me);
    % Correct for ties
    n_ties = sum (test_stat == sample_stat);
    u = rand (n_ties + 1, 1);
    % Estimate survival function
    g = (sum (test_stat < sample_stat) + sum (u(end) >= u(1:n_ties))) / n_mc;
end

% Constraints for the maximum entropy distribution
function [lambda1 lambda2 correlation] = constraints (p)
    [n1, n2] = size (p);
    % Marginal distributions
    margin1 = sum (p, 2)';
    margin2 = sum (p, 1);
    % Expectations
    lambda1 = dot (margin1, 0:n1-1);
    lambda2 = dot (margin2, 0:n2-1);
    % Variances
    sigma1 = dot (margin1, ((0:n1-1)-lambda1).^2);
    sigma2 = dot (margin2, ((0:n2-1)-lambda2).^2);
    % Correlation coefficient
    if (sigma1 == 0 || sigma2 == 0)
        correlation = 0;
    else
        correlation = (sum (sum (p .* ((0:n1-1)' * (0:n2-1)))) - lambda1*lambda2) ./ sqrt (sigma1*sigma2);
    end
end

% Entropy of the probability mass function p
function h = entropy (p)
    h = p .* log2 (p);
    h(p <= 0) = 0;
    h = -sum (h);
end
