% Copyright (C) 2012  Arno Onken
%
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% -------------------------------------------------------------------------
% Searches the maximum value of the function f using simulated annealing
% (Kirkpatrick, Gelatt and Vecchi, Science, 1983).
%
% Arguments:
%  f      - Objective function
%  x_init - Initial point
%  lb     - Lower bound of x
%  ub     - Upper bound of x
%  t_init - Initial temperature
%
% Returns:
%  maxval - Maximum function value
% -------------------------------------------------------------------------
function maxval = sup (f, x_init, lb, ub, t_init)
    % Initial temperature
    t = t_init;
    % Number of differences to average over
    n_cond = 10;
    % Initial point
    x = x_init;
    val_x = f (x);
    % Initial maximum value
    maxval = val_x;
    % Array with differences
    diff = ones (n_cond, 1);
    % Circular index for diff
    index = 1;
    % Optimization loop
    while (mean (abs (diff)) > 1e-6)
        % Decrease temperature
        t = 0.95 * t;
        % Pick uniform random direction from the unit hypersphere
        % (Marsaglia, The Annals of  Mathematical Statistics, 1972)
        v = normrnd (zeros (size (x)), 1);
        v = v ./ sqrt (sum (v .^ 2));
        % Go t into that direction
        x_prop = x + t * v;
        % Out of bounds?
        if (any (x_prop < lb | x_prop > ub))
            continue;
        end
        val_prop = f (x_prop);
        delta = val_x - val_prop;
        % Accept point?
        if (delta < 0 || rand < 1 / (1 + exp (delta/t)))
            x = x_prop;
            val_x = val_prop;
            % New maximum value?
            if (val_x > maxval)
                maxval = val_x;
            end
            diff(index) = delta;
        else
            diff(index) = 0;
        end
        % Set next circular index
        index = mod (index, n_cond) + 1;
    end
end